import 'dart:convert';
import 'dart:html' as html;
import 'dart:typed_data';

import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter_markdown/flutter_markdown.dart';
import 'package:google_fonts/google_fonts.dart';
import 'package:remove_emoji/remove_emoji.dart';
import 'package:markdown/markdown.dart' as md;
import 'dart:ui' as ui;
import 'package:url_launcher/url_launcher.dart';
import 'package:dotted_border/dotted_border.dart';
import 'package:drop_zone/drop_zone.dart';

abstract class Parser {
  Widget parse(Map block, {Options options, BuildContext context});
}

class Options {
  const Options({this.margin = const EdgeInsets.all(0), this.disabled = false});

  final EdgeInsets margin;
  final bool disabled;
}

class ButtonParser implements Parser {
  const ButtonParser();

  @override
  Widget parse(Map block, {Options options = const Options(), BuildContext context}) {
    //var action = block["action"] != null ? block["action"] : {};
    var text = block["text"];
    var hasEmoji = text["emoji"];
    var innerType = text["type"];
    var innerText = text["text"];
    if (hasEmoji != null && hasEmoji == false) {
      innerText = RemoveEmoji().removemoji(innerText);
    }

    Widget inner = Container();
    if (innerType == "plain_text") {
      inner = Container(
        child: Text(innerText,
            style: GoogleFonts.getFont('Open Sans', fontSize: 14, fontWeight: FontWeight.w600, color: Colors.black)),
        padding: EdgeInsets.fromLTRB(8, 6, 8, 6),
        decoration: BoxDecoration(
            border: Border.all(width: 0.5, color: Colors.grey.shade600),
            borderRadius: BorderRadius.all(Radius.circular(4))),
      );
    } else if (innerType == "mrkdwn") {
      inner = MarkdownBody(
        data: innerText,
        extensionSet: md.ExtensionSet(
          md.ExtensionSet.gitHubFlavored.blockSyntaxes,
          [md.EmojiSyntax(), ...md.ExtensionSet.gitHubFlavored.inlineSyntaxes],
        ),
        onTapLink: (text, href, title) {
          launch(href);
        },
      );
    }

    return Container(
        margin: options.margin,
        child: TextButton(
          onPressed: () {},
          child: inner,
          style: ButtonStyle(overlayColor: MaterialStateProperty.all(Colors.grey.shade100)),
        ));
  }
}

class SpacerParser implements Parser {
  const SpacerParser();

  @override
  Widget parse(Map block, {Options options = const Options(), BuildContext context}) {
    return Spacer();
  }
}

class SectionParser implements Parser {
  const SectionParser();

  @override
  Widget parse(Map block, {Options options = const Options(), BuildContext context}) {
    var tb = block["text"];
    var text = tb["text"];

    if (tb != null) {
      var hasEmoji = tb["emoji"];
      if (hasEmoji != null && hasEmoji == false) {
        text = RemoveEmoji().removemoji(text);
      }
      if (tb["type"] == "plain_text") {
        return Text(
          text,
          textAlign: TextAlign.left,
          style: GoogleFonts.getFont('Open Sans', fontSize: 14, fontWeight: FontWeight.w400, color: Colors.black),
        );
      } else if (tb["type"] == "mrkdwn") {
        return Container(
          child: MarkdownBody(
            data: text,
            extensionSet: md.ExtensionSet(
              md.ExtensionSet.gitHubFlavored.blockSyntaxes,
              [md.EmojiSyntax(), ...md.ExtensionSet.gitHubFlavored.inlineSyntaxes],
            ),
            styleSheet: MarkdownStyleSheet.fromTheme(Theme.of(context)).copyWith(
                p: GoogleFonts.getFont('Open Sans',
                    fontSize: 14, fontWeight: FontWeight.w400, color: Colors.black, height: 1.4)),
            onTapLink: (text, href, title) {
              launch(href);
            },
          ),
        );
      }
    }
    return Container();
  }
}

class DividerParser implements Parser {
  const DividerParser();

  @override
  Widget parse(Map block, {Options options = const Options(), BuildContext context}) {
    return Divider(
      color: Colors.grey.shade500,
      height: 12,
      thickness: 0.5,
    );
  }
}

class HeaderParser implements Parser {
  const HeaderParser();

  @override
  Widget parse(Map block, {Options options = const Options(), BuildContext context}) {
    var text = block["text"];
    var innerType = text["type"];
    var innerText = text["text"] != null ? text["text"] : "";
    var hasEmoji = text["emoji"];
    if (hasEmoji != null && hasEmoji == false) {
      innerText = RemoveEmoji().removemoji(innerText);
    }

    Widget inner = Container();
    if (innerType == "plain_text") {
      inner = Container(
          alignment: Alignment.centerLeft,
          padding: EdgeInsets.fromLTRB(0, 8, 0, 4),
          child: Text(innerText,
              style:
                  GoogleFonts.getFont('Open Sans', fontSize: 18, fontWeight: FontWeight.w700, color: Colors.black87)));
    } else if (innerType == "mrkdwn") {
      inner = MarkdownBody(
        data: innerText,
        extensionSet: md.ExtensionSet(
          md.ExtensionSet.gitHubFlavored.blockSyntaxes,
          [md.EmojiSyntax(), ...md.ExtensionSet.gitHubFlavored.inlineSyntaxes],
        ),
        onTapLink: (text, href, title) {
          launch(href);
        },
      );
    }

    return inner;
  }
}

class ImageParser implements Parser {
  const ImageParser();

  @override
  Widget parse(Map block, {Options options = const Options(), BuildContext context}) {
    var title = block["title"];
    var url = block["image_url"];
    var innerType = title != null ? title["type"] : "";

    var _image = html.ImageElement(src: url);
    // ignore: undefined_prefixed_name
    ui.platformViewRegistry.registerViewFactory(
      "_image",
      (int viewId) => _image,
    );

    Widget inner = Container();
    if (innerType == "plain_text") {
      var header = title["text"] != null ? title["text"] : "";

      inner = Container(
          alignment: Alignment.centerLeft,
          padding: EdgeInsets.fromLTRB(0, 8, 0, 4),
          child: Text(header,
              style: GoogleFonts.getFont('Open Sans',
                  fontSize: 14, fontWeight: FontWeight.w400, color: Colors.grey.shade600)));
    } else if (innerType == "mrkdwn") {
      inner = MarkdownBody(
        data: title["text"],
        extensionSet: md.ExtensionSet(
          md.ExtensionSet.gitHubFlavored.blockSyntaxes,
          [md.EmojiSyntax(), ...md.ExtensionSet.gitHubFlavored.inlineSyntaxes],
        ),
        onTapLink: (text, href, title) {
          launch(href);
        },
      );
    }

    return Container(
        child: Column(
      children: [
        inner,
        Container(
            alignment: Alignment.centerLeft,
            child: SizedBox(
                width: _image.width.toDouble(),
                height: _image.height.toDouble(),
                child: HtmlElementView(viewType: "_image"))),
      ],
    ));
  }
}

class DropWidget extends StatefulWidget {
  final void Function(Map<html.File, Uint8List>) onDropCompleted;
  final Options options;

  DropWidget({this.onDropCompleted, this.options = const Options()});

  @override
  createState() => DropWidgetState();
}

class DropWidgetState extends State<DropWidget> {
  Map<html.File, Uint8List> files = {};
  bool reading = false;
  bool dragged = false;

  @override
  Widget build(BuildContext context) {
    bool disabled = files.length > 0 || widget.options.disabled;

    return GestureDetector(
        onTap: disabled
            ? null
            : () {
                if (files.length > 0) return;

                html.InputElement uploadInput = html.FileUploadInputElement();
                uploadInput.click();

                uploadInput.onChange.listen((e) {
                  final files = uploadInput.files;
                  onDrop(files);
                  setState(() {
                    dragged = true;
                  });
                });
              },
        child: Container(
            width: 450,
            height: 250,
            padding: EdgeInsets.all(12),
            color: dragged ? Colors.blueGrey.shade100 : Colors.grey.shade200,
            child: Stack(
              children: [
                DottedBorder(
                  color: disabled ? Colors.grey.shade500 : Colors.grey.shade800,
                  dashPattern: dragged ? [8, 0] : [8, 4],
                  strokeCap: StrokeCap.round,
                  strokeWidth: 1,
                  child: DropZone(
                    onDragEnter: disabled
                        ? null
                        : () {
                            setState(() {
                              dragged = true;
                            });
                          },
                    onDragExit: disabled
                        ? null
                        : () {
                            setState(() {
                              dragged = false;
                            });
                          },
                    onDrop: disabled
                        ? null
                        : (List<html.File> files) {
                            onDrop(files);
                          },
                    child: Container(
                      child: Center(
                        child: Column(
                          mainAxisAlignment: MainAxisAlignment.center,
                          children: [
                            Icon(
                              Icons.upload_file,
                              size: 72,
                              color: disabled ? Colors.grey.shade500 : Colors.blueGrey,
                            ),
                            Container(
                                margin: EdgeInsets.only(top: 12, bottom: 12),
                                child: Row(
                                  mainAxisAlignment: MainAxisAlignment.center,
                                  children: getText(),
                                ))
                          ],
                        ),
                      ),
                    ),
                  ),
                ),
                Positioned(
                  right: 4,
                  top: 4,
                  child: SizedBox(
                      width: 32,
                      child: TextButton(
                        onPressed: widget.options.disabled
                            ? null
                            : () {
                                setState(() {
                                  files = {};
                                  reading = false;
                                  dragged = false;
                                });
                              },
                        child: Container(child: Icon(Icons.close)),
                        style: ButtonStyle(
                            foregroundColor:
                                MaterialStateProperty.all(disabled ? Colors.grey.shade500 : Colors.amber.shade700),
                            backgroundColor: MaterialStateProperty.all(Colors.transparent),
                            overlayColor: MaterialStateProperty.all(Colors.transparent),
                            padding: MaterialStateProperty.all(EdgeInsets.zero)),
                      )),
                )
              ],
            )));
  }

  List<Widget> getText() {
    bool disabled = files.length > 0 || widget.options.disabled;

    if (reading) {
      return [
        Text("Preparing files...",
            style: GoogleFonts.getFont('Open Sans', fontSize: 16, fontWeight: FontWeight.w700, color: Colors.blueGrey)),
      ];
    }

    if (files.length == 0) {
      return [
        Text("Choose a file",
            style: GoogleFonts.getFont('Open Sans',
                fontSize: 16, fontWeight: FontWeight.w700, color: disabled ? Colors.grey.shade500 : Colors.blueGrey)),
        Text(" or drag it here.",
            style: GoogleFonts.getFont('Open Sans',
                fontSize: 16, fontWeight: FontWeight.w400, color: disabled ? Colors.grey.shade500 : Colors.blueGrey))
      ];
    }

    List<Widget> names = [
      Text("Done!",
          style: GoogleFonts.getFont('Open Sans', fontSize: 16, fontWeight: FontWeight.w700, color: Colors.blueGrey))
    ];

    for (var file in files.keys) {
      names.add(Text(file.name,
          style: GoogleFonts.getFont('Open Sans', fontSize: 14, fontWeight: FontWeight.w700, color: Colors.blueGrey)));
    }

    return [
      Column(
        children: names,
      )
    ];
  }

  Future<void> onDrop(List<html.File> fileArr) async {
    setState(() {
      reading = true;
    });

    Map<html.File, Uint8List> _files = {};
    for (html.File file in fileArr) {
      html.FileReader reader = html.FileReader();
      reader.readAsArrayBuffer(file);
      await reader.onLoadEnd.first;
      _files[file] = reader.result;
    }

    if (widget.onDropCompleted != null) widget.onDropCompleted(_files);

    setState(() {
      reading = false;
      files = _files;
    });
  }
}

class DragDropParser implements Parser {
  Map<html.File, Uint8List> files = {};
  bool reading = false;

  @override
  Widget parse(Map block, {Options options = const Options(), BuildContext context}) {
    return DropWidget(
      onDropCompleted: this.onDropCompleted,
      options: options,
    );
  }

  void onDropCompleted(Map<html.File, Uint8List> files) {}
}
